from math import log10

def create_HMM(cpg, non_cpg):
  HMM = {
    0: {
        'emissions': {
          'A': 1,
          'T': 1,
          'C': 1,
          'G': 1,
        },
        
        'transitions': {
          0: 0.8,
          1: 0.2
        },
        
        'label': '-'
      },
    1: {
        'emissions': {
          'A': 1,
          'T': 1,
          'C': 1,
          'G': 1,
        },
        
        'transitions': {
          0: 0.1,
          1: 0.9
        },
        
        'label': '+'
      }
    }
    
  for sequence in cpg:
    for symbol in sequence:
      HMM[1]['emissions'][symbol] += 1
      
  for sequence in non_cpg:
    for symbol in sequence:
      HMM[0]['emissions'][symbol] += 1 
      
  cpg_count = 0
  
  for (symbol, num) in HMM[1]['emissions'].items():
    cpg_count += num
    
  for (symbol, num) in HMM[1]['emissions'].items():
    HMM[1]['emissions'][symbol] = num / cpg_count
    
  non_cpg_count = 0
  
  for (symbol, num) in HMM[0]['emissions'].items():
    non_cpg_count += num
    
  for (symbol, num) in HMM[0]['emissions'].items():
    HMM[0]['emissions'][symbol] = num / non_cpg_count
    
  return HMM
  
  
def viterbi(HMM, sequence):
  n = len(HMM)
  m = len(sequence)
  
  viterbi_matrix = [[0 for j in range(m)] for i in range(n)]
  backtrack = [[0 for j in range(m)] for i in range(n)]
  
  for j in range(m):
  
    c_j = sequence[j]
  
    if j == 0:
      p_i_1 = 0
      a = 1/n
      
      for i in range(n):
        e_j = HMM[i]['emissions'][c_j]
        viterbi_matrix[i][j] = p_i_1 + log10(a) + log10(e_j)
        backtrack[i][j] = None
      
    else:
      for i in range(n):
        
        p_i = float('-inf')
        
        for k in range(n):
          p_i_1 = viterbi_matrix[k][j-1]
          a = HMM[k]['transitions'][i]
          e_j = HMM[i]['emissions'][c_j]
          
          p_tmp = p_i_1 + log10(a) + log10(e_j)
          
          if p_tmp > p_i:
            p_i = p_tmp
            backtrack[i][j] = (k, j-1)
            
        viterbi_matrix[i][j] = p_i

  j = m - 1
  i = 0
  max_p = float('-inf')
  
  for k in range(n):
    if viterbi_matrix[k][j] > max_p:
      max_p = viterbi_matrix[k][j]
      i = k
      
  print(max_p)
      
  path = HMM[i]['label']
      
  while backtrack[i][j] != None:
    path = HMM[i]['label'] + path
    (i,j) = backtrack[i][j]
      
  return path
        
        
    

  
cpg = [
  'CGCGGCGGCCGATGCGCCTGACGGC',
  'GCCGATTAGCTAAGCGTAGGCTAAG'
  'CCGCGGCGGCATTAGCGCGTACGGA'
  'TCCGGACTGGCGGACCGGCATCGAACA'
  'CGGATGCCGGCAGCGGGCGGCG'
]

non_cpg = [
  'TAATCTATTATGCTAGGCTGATATT',
  'TTAGCTGATTATATGCAGTGTATTA'
  'TATAGGCTTCTAAGGCCAAGGATAA'
  'CATTGATACAGTAAATAGTATGCAT'
  'CTTGAAACATTTAGATATATCATTA'
]

HMM = create_HMM(cpg, non_cpg)

print(HMM)


sequence = 'GTCTGTACGCGTAGGCTATGTGCAAAAAAAAAAGCCGCGCGCGCGCGCGCGC'

print(viterbi(HMM, sequence))
print(sequence)
